/*  Initial tests for Greenspan/Barnes/Hut programs	*/
/*  Source file:  gbhtst.cs				*/

/* Local header files are included after declarations.	*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <cscomm.h>

/*  Definitions corresponding to declarations in gbh.hs		*/

 int	    N,	/* Max particles			*/
	   NT,	/* Dimension of tree, usually = 2*N	*/
	    T,	/* Number of levels in tree generation	*/
	Nodes,	/* Number of nodes in the tree		*/
	   HT;	/* Height of tree			*/

 shape []Tree;	/* Will be [NT]Tree	*/

 int:Tree
	*pParent,	/* Parent cell		*/
	*pTnext,	/* Next node in search	*/
	*pChild0,	/* 0-th child cell	*/
	*pChldrn;	/* Identifies children	*/

/* Dereferenced forms for convenience	*/
#define Parent (*pParent)
#define Tnext  (*pTnext)
#define Child0 (*pChild0)
#define Chldrn (*pChldrn)

 shape []Particles;	/* Will be [N]Particles	*/

 float:Particles
		*X0,		/* Will be X0[3] */
		*V0,		/* Will be V0[3] */
		*X1,		/* Will be X1[3] */
		*V1,		/* Will be V1[3] */
		*Ac,		/* Will be Ac[3] */
		*pgm;		/* G*mass	 */

 int:Particles    *i,		/* Particle coordinate	*/
	      *pcell;		/* Tree leaf corresponding to particle */

#define gm   (*pgm)
#define Ip   (*i)
#define cell (*pcell)

float Ba[3], Bb[3];		/* Bounding box/cube */

#define moo (0x40000000)	/* -infinity */

/* Parallel data sizes */

int pfsz, pvsz, pisz, tisz;
/* Particle float, vector, int, Tree int */


#include "accel1.hs"
#include "accel2.hs"
#include "acdiag.hs"
#include "fill.hs"
#include "tree.hs"
#include "compare.hs"
#include "nb1d.hs"

/* Additional problem parameters and data */

float G;	/* Constant of gravitation	*/
int  PX,	/* Print control		*/
     NB;	/* Number of bodies so far	*/


/*  Allocate shapes and global parallel variables.	*/

void Hcreate_globals()
 { NT = 2*N;				/* Always enough */
   allocate_shape(&Particles, 1, N);
   allocate_shape(&Tree, 1, NT);
   pfsz = boolsizeof(float:Particles);  pvsz = 3*pfsz;
   pisz = boolsizeof(int:Particles);
   tisz = boolsizeof(int:Tree);
   /* Float vectors[3] */
   X0 = palloc(Particles, pvsz);  X1 = palloc(Particles, pvsz);
   V0 = palloc(Particles, pvsz);  V1 = palloc(Particles, pvsz);
   Ac = palloc(Particles, pvsz);
   /* Floats */
   pgm  = palloc(Particles, pfsz);
   /* Ints  */
   i = palloc(Particles, pisz);    pcell = palloc(Particles, pisz);
   pParent = palloc(Tree, tisz);   pTnext  = palloc(Tree, tisz);
   pChild0 = palloc(Tree, tisz);   pChldrn = palloc(Tree, tisz);
 }

FILE *datap;      /* Stream for reading data */

/* Load inital positions, velocities, also masses, gravitational const. */
/* Leaves data for Pi in processor [i].					*/

void init()
 { int k;  float a, b, c;
   for ( k = 0; k < NB; k++ )
    { fscanf(datap, "%e%e%e", &a, &b, &c);
      [k]X0[0] = a;  [k]X0[1] = b;  [k]X0[2] = c;
    };
   for ( k = 0; k < NB; k++ )
    { fscanf(datap, "%e%e%e", &a, &b, &c);
      [k]V0[0] = a;  [k]V0[1] = b;  [k]V0[2] = c;
    };
   for ( k = 0; k < NB; k++ )
     { fscanf(datap, "%e", &a);  [k]*pgm = a; };
   fscanf(datap, "%e", &G);
 }


/* Print current state of system according to PX.  See comments for	*/
/* main program for PX settings.					*/

void print_state(float:Particles X[3], float:Particles V[3],
		 float:Particles *m			  )
 { int k, np;
   if ( (PX >= 0) && (PX < NB) ) np = PX; else np = NB;
   if ( np > 0 )
    { printf(
  "\n        x         y         z         x         y         z       m\n");
      for ( k = 0; k < np; k++ )
      { printf("%2d %9.5f %9.5f %9.5f %9.5f %9.5f %9.5f %9.2e\n",
                 k, [k]X[0], [k]X[1], [k]X[2],
                    [k]V[0], [k]V[1], [k]V[2], [k]*m);
      }
   }
 }


/* Insert nx asteroids at radius ar with total mass am, random z	*/
/* components in [-az:az]..  This assumes that an orbitof radius 1	*/
/*  about theorigin corresponds to velocity 2*pi (au-earth-yr units).	*/

#define TWOPI 6.283185307f
#define two31 2147483648.0

void mkast(int nx, float ar, float am, float az)
 { float:Particles sx, cx, arg;
   float spd;  int nn;
   nn = NB + nx;
   spd = TWOPI/sqrt(ar);
   printf("Making %d asteroids, r = %9.5f, spd = %9.5f, az = %9.5f\n",
		  nx,		     ar,	  spd,        az      );
   printf("                     tot mass = %9.2e\n",  am);
   with ( Particles )
   { where ( (NB <= Ip) & (Ip < nn) )
      { arg = (TWOPI*(Ip - NB))/nx;
        sx = sin(arg);   cx = cos(arg);
        X0[0] =  cx*ar;  V0[0] = (-sx)*spd;
        X0[1] =  sx*ar;  V0[1] =  cx*spd;
        X0[2] =  az*(2.0*prand()/two31 - 1.0);
        V0[2] =  0.0;
        gm = am/nx;
      }}  /* end where, end with (Particles) */
   NB = nn;
 }


/* Main program:  reads problem and control information from data file	*/
/* named by user and performs calculation.				*/
/* Print control:							*/
/*	  0 - just performance data					*/
/*	< 0 - all							*/
/*	> 0 - number given						*/

void main()
 { int ns, np, u;  float dt, eps, ar, am, az;
   char fnm[50];

   printf("Enter N for [N]Particles, [2*N]Tree > ");
   scanf("%d", &N);
   Hcreate_globals();
   with ( Particles ) Ip = pcoord(0);
   /* Loop for test cases */
   while ( printf("Data file (or q)> "),
	   scanf("%s", fnm),
	   ( fnm[0] != 'q' ) || ( fnm[1] > 0 ) )
    { /* Explores one initial state */
      datap = fopen(fnm, "r");
      if ( datap == NULL )
        { printf("Can't open file.\n"); continue; }
      printf("Print ctl> ");  scanf("%d", &PX);
      fscanf(datap, "%d", &NB);
      printf("Initialize....\n");
      init();    /* Read data for individual bodies */
      /* Read "asteroid" descriptions */
      while ( fscanf(datap, "%d", &ns), ns > 0)
       { fscanf(datap, "%e%e%e", &ar, &am, &az);  mkast(ns, ar, am, az); }
      fscanf(datap, "%d%d%e%e", &ns, &np, &dt, &eps);
      fclose(datap);
      printf("\n   NB = %d    G = %13.6e", NB, G);
      with (Particles) everywhere
       { float:Particles Xc[3], Aref[3];
	 float:Tree Xp0[3], Xp1[3], Tm, Th;
         int acyc, cube, dtl;  float theta, beta;  double ndx, nda;

	 printf("\nX0, V0, Unscaled masses:");
	 print_state(X0, V0, pgm);
	 gm *= G;			/* Scale masses */
       while ( printf("delta-t ( < 0 to quit ) > "),  scanf("%f", &dt),
	       ( dt >= 0.0f ) )
       { for ( u = 0;  u < 3; u++ )
	  { Ac[u] = 0.0;
	    V1[u] = V0[u] + dt*Ac[u];
	    X1[u] = X0[u] + 0.5f*dt*(V0[u] + V1[u]);
	    Xc[u] = 0.5f*(X0[u] + X1[u]);
	  }
	 printf("\nX1, Xc, Scaled masses:");
	 print_state(X1, Xc, pgm);
         /* acsmain(Aref);	/* Calculate reference acceleration */
       while ( cube = Toption(),
	       ( cube >= 0 ) )
       { int prp, acv;
	 printf("Particle representative { 0 | c | 1 } > ");
         scanf("%s", fnm);
         prp = (fnm[0] == '0') ? 0 : (fnm[0] == '1') ? 1 : 2;
	 CM_timer_clear(1);  CM_timer_start(1);
	 cell = TreeTop((prp == 0) ? X0 : (prp == 1) ? X1 : Xc, cube);
	 CM_timer_stop(1);   
         printf("Tree construction:");  CM_timer_print(1);
	 CM_timer_clear(2);  CM_timer_start(2);
         Tfill(Xp0, Xp1, &Tm, &Th);
         CM_timer_stop(2);
	 printf("Tree filling:");  CM_timer_print(2);
	 printf("Details? { -1 | 0 | 1 } > ");  scanf("%d", &dtl);
         if ( dtl == -1 ) Tsummary();
	  else { Tdetail(dtl, Xp0, Xp1, &Tm, &Th);  Pcells(&cell); }
	 /* Calculate Ac for various theta, beta, accel versions */
       while ( printf("theta, beta ( < 0 to quit ) > "),
	       scanf("%f%f", &theta, &beta),
	       ( theta >= 0.0f ) )
      { while ( printf("Accel version (<= 0 to quit) > "),
		scanf("%d", &acv),
		(acv > 0) )
       { CM_timer_clear(3);  CM_timer_start(3);
	 acyc = (acv == 1)
		 ? Accel1(Xp0, Xp1, &Tm, &Th, theta, beta, &ndx, &nda)
	      : (acv == 2)
		 ? Accel2(Xp0, Xp1, &Tm, &Th, theta, beta, &ndx, &nda)
		 : Accel3(Xp0, Xp1, &Tm, &Th, theta, beta, &ndx, &nda);
	 CM_timer_stop(3);
	 if ( acv == 1) for ( u = 0;  u < 3; u++ ) Aref[u] = Ac[u];
	 printf("Acceleration:");  CM_timer_print(3);
	 printf("cube = %d   theta = %8.5f   beta = %8.5f   delta-t = %8.5f\n",
		       cube,         theta,          beta,              dt   );
	 Asummary(acyc, ndx, nda,
			(prp == 0) ? '0' : (prp == 1) ? '1' : 'c');
       }	/* end version loop */
	 compare(Aref, Ac);
	 printf("Ac, Aref, Scaled masses:");
	 print_state(Ac, Aref, pgm);
      }		/* end theta-beta loop	*/
       }	/* end cube loop	*/
       }	/* end delta-t loop	*/
       }	/* end with (Particles)	*/
    }  /* end while (file)	*/
 }
